/*-
 * dlpi_wire.cc
 *
 * Implementation of Wire for dlpi
 */
/*
 * This has so far been tested only on Solaris-SPARC. It should also work
 * on Solaris-x86 if you can figure out the proper device name.  It should
 * also work on AIX, SINIX, HPUX and others provided you can persuade it
 * to see the relevnt #ifdefs below (though a lot of further hacking may
 * be required for HPUX).  the overall strategy (though not the detailed
 * coding) was culled from the libpcap sources (including the various
 * #ifdefs, as mentioned).
 */
#include "config.h"

#if HAVE_DLPI
/* For documentation on DLPI, see http://www.opengroup.org/onlinepubs/9638599
*/

#include <cstring>
#include <stdexcept>
#include <cerrno>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/select.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/dlpi.h>
#include <sys/stream.h>
#include <stropts.h>
#include <errno.h>
#include <stdio.h>

#include "nslu2_upgrade.h"

namespace NSLU2Upgrade {
	/* The basic class implemented to transmit and receive packets over the
	 * wire.
	 */
	class DlpiWire : public Wire {
	public:
		DlpiWire(int fd, const char *device, const unsigned char *mac,
				const unsigned char address[6]) :
			file(fd), broadcast(address == 0) {

			/* The 255 gives the ethernet hardware broadcast address,
			 * set this if a host address is provided.  The packet
			 * header is:
			 *
			 * target MAC      [6 bytes] target or broadcast
			 * originating MAC [6 bytes] MAC of this device
			 * protocol        [2 bytes] 0x8888 (big endian)
			 */
			if (address)
				std::memcpy(header, address, 6);
			else
				std::memset(header, 255, 6);
			header[12] = NSLU2Protocol::UpgradeProtocol >> 8;
			header[13] = NSLU2Protocol::UpgradeProtocol;

			/* This is set just in case of a call to LastAddress before
			 * Receive has succeeded - the result will be all 0's
			 */
			std::memset(source, 0, sizeof source);

			/* And copy the MAC address into the header. */
			std::memcpy(header+6, mac, 6);
		}

		virtual ~DlpiWire() {
			close(file);
		}

		/* Throws SendError on a fatal error. */
		virtual void Send(const void *packet, size_t length) {
			if (length > 1540-14)
				throw std::logic_error("packet too large");

			/* Set no flags (0) - we block on the transmit if
			 * required, the file is *not* set O_NONBLOCK.
			 */
			char buffer[1540];
			std::memcpy(buffer, header, 14);
			std::memcpy(buffer+14, packet, length);

			int   len(length+14);
			do {
				struct strbuf data = {0, len, buffer};
				if (putmsg(file, (struct strbuf *)NULL, &data, 0) != 0 &&
					errno != EINTR)
						throw SendError(errno);
				len -= data.len;
			} while (len > 0);
		}

		/* Receive throws ReceiveError on a fatal error and must update
		 * size with the received packet size.  0 must be used to
		 * indicate failure to receive a packet (and this must not
		 * be fatal).  If timeout is greater than 0 the implementation
		 * should wait that number of microseconds until a packet is
		 * received or the timeout has expired (in which case a size
		 * of 0 must be returned).
		 */
		virtual void Receive(void *buffer, size_t &size, unsigned long timeout) {

			char bigbuffer[1540];

			struct timeval tv;
			int fds;
			tv.tv_sec = timeout >> 20;
			tv.tv_usec = timeout & 0xfffff;
			if (tv.tv_usec >= 1000000)
				++tv.tv_sec, tv.tv_usec = 0;
			do {
				struct strbuf data = {size, 0, bigbuffer};
				fd_set readfds;
				int flags = 0;

				/* See if there is anything to read... */
				do {
					FD_ZERO(&readfds);
					FD_SET(file, &readfds);
					int fds = select(file+1, &readfds, NULL, NULL, &tv);
					if (fds >0)
						break;
					else if (fds == 0) {
						size = 0;
						return;
					} else if (fds < 0 && errno != EINTR)
						throw ReceiveError(errno);
				} while (1);
				/* There is something to read... */
				if (getmsg(file, (struct strbuf *)NULL, &data, &flags) < 0 &&
					errno != EINTR)
						throw ReceiveError(errno);
				if (data.len > 14 &&
					(broadcast || std::memcmp(data.buf+6, header, 6) == 0)) {
						/* this is a response to our Send */
						std::memcpy(source, data.buf+6, 6);
						size = data.len-14;
						std::memcpy(buffer, data.buf+14, size);
						return;
				}
			} while (1);		
		}

		/* Return the address of the last received packet.  This is
		 * an NSLU2 so the address is a 6 byte Ethernet hardware
		 * address.
		 */
		virtual void LastAddress(unsigned char address[6]) {
			std::memcpy(address, source, 6);
		}

	private:
		int     file;           /* file descriptor */
		char    header[14];     /* Packet header. */
		char    source[6];      /* Source of last *received* packet. */
		bool    broadcast;
	};

	/* Class to set and reset the user id to the effective uid. */
	class EUID {
	public:
		EUID(int uid) : euid(::geteuid()) {
			if (uid != -1 && ::seteuid(uid) != 0)
				throw WireError(errno);
		}

		~EUID() {
			::seteuid(euid);
		}

	private:
		::uid_t euid;
	};

};

#define	MAXDLBUF	8192
#define DL_ATTACH_ACK DL_OK_ACK
#define DL_ATTACH_ACK_SIZE DL_OK_ACK_SIZE
#define DL_PROMISCON_ACK DL_OK_ACK
#define DL_PROMISCON_ACK_SIZE DL_OK_ACK_SIZE

/*
 * There are 3 variants of the REQUEST Macro, according to how many fields
 * are to be set.
 */
#define REQUEST2F(lower,upper,field,field2) { \
	dl_ ## lower ## _req_t req; \
	memset((char *)&req, 0, sizeof(req)); \
	req.dl_primitive = DL_ ## upper ## _REQ; \
	req.dl_ ## field; \
	req.dl_ ## field2; \
	ctl.maxlen = 0; \
	ctl.len = sizeof(req); \
	ctl.buf = (char *)&req; \
	if (putmsg(fd, &ctl, (struct strbuf *)NULL, 0) < 0) \
		throw WireError(errno, "DLPI " #lower " request failure"); \
}
#define REQUESTF(lower,upper,field) { \
	dl_ ## lower ## _req_t req; \
	memset((char *)&req, 0, sizeof(req)); \
	req.dl_primitive = DL_ ## upper ## _REQ; \
	req.dl_ ## field; \
	ctl.maxlen = 0; \
	ctl.len = sizeof(req); \
	ctl.buf = (char *)&req; \
	if (putmsg(fd, &ctl, (struct strbuf *)NULL, 0) < 0) \
		throw WireError(errno, "DLPI " #lower " request failure"); \
}
#define REQUEST(lower,upper) { \
	dl_ ## lower ## _req_t req; \
	req.dl_primitive = DL_ ## upper ## _REQ; \
	ctl.maxlen = 0; \
	ctl.len = sizeof(req); \
	ctl.buf = (char *)&req; \
	if (putmsg(fd, &ctl, (struct strbuf *)NULL, 0) < 0) \
		throw WireError(errno, "DLPI " #lower " request failure"); \
}
#define ACK(lower,upper,dlp) { \
	ctl.maxlen = MAXDLBUF; \
	ctl.len = 0; \
	ctl.buf = (char *)dlp; \
	int flags = 0; \
	if (getmsg(fd, &ctl, (struct strbuf *)NULL, &flags) < 0) \
		throw WireError(errno, "DLPI " #lower " acknowledge failure"); \
	if (dlp->dl_primitive != DL_ ## upper ## _ACK) \
		if (dlp->dl_primitive == DL_ERROR_ACK && \
		    dlp->error_ack.dl_errno == DL_SYSERR) \
			throw WireError(dlp->error_ack.dl_unix_errno, \
					"DLPI " #lower "acknowledge syserr"); \
		else { \
			snprintf(Errbuf, ERRBUF_SIZE, \
				 "DLPI " #lower " acknowledge error #%d", \
				 dlp->error_ack.dl_errno); \
			throw WireError(0, Errbuf); \
		} \
	if (ctl.len < DL_ ## upper ## _ACK_SIZE) \
		throw WireError(0, "DLPI " #lower " ACK too small"); \
}

#define ERRBUF_SIZE 256

#ifdef HAVE_DEV_DLPI
#include "dev-dlpi.c"
/*
 * This arises with (some?) HPUX systems. Note that the #included file is
 * taken from libpcap, and is subject to a BSD-like Copyright.
 */
#endif

/*
 * Make a new wire, which may be deleted with delete.  The address should
 * be a value (null terminated this time) returned by LastAddress, if NULL
 * the Wire will broadcast.
 */
NSLU2Upgrade::Wire *NSLU2Upgrade::Wire::MakeWire(const char *device,
		const unsigned char *mac, const unsigned char *address, int uid) {
	/* This is used to store the error passed to throw. */
	static char Errbuf[ERRBUF_SIZE];

	int fd;
	int unit;
	struct strbuf ctl;
	char longbuf[MAXDLBUF];
	union DL_primitives *ackbuf = (DL_primitives *)longbuf;
	static unsigned char macBuffer[6];
	EUID euid(uid);
	char dname[12], dname2[12];
	char *cp;
		
	if (device == NULL)
		throw WireError(0, "no default device avalable");
/*
 * Split a device name into a device type name and a unit number
*/
	strcpy(dname, device);
	cp = dname + strlen(dname) - 1;
	if (!isdigit(*cp))
		throw WireError(0, "missing unit number");
	unit = *cp - '0';
	*cp = '\0';
#ifdef HAVE_DEV_DLPI
	if ((fd = open("/dev/dlpi", O_RDWR)) < 0)
		throw WireError(errno);
	if ((unit = get_dlpi_ppa(fd, dname, unit, Errbuf)) < 0)
			throw WireError(0, Errbuf);
#else
	snprintf(dname2, sizeof(dname2), "/dev/%s", dname);
	if ((fd = open(dname2, O_RDWR)) < 0) {
		dname2[strlen(dname2)] = unit + '\0';
		if (errno == ENOENT && (fd = open(dname2, O_RDWR)) >= 0)
			unit = 0;
		else
			throw WireError(errno);
	}
#endif

	REQUEST(info,INFO)
	ACK(info,INFO,(ackbuf))
	if (ackbuf->info_ack.dl_provider_style == DL_STYLE2) {
		REQUESTF(attach,ATTACH,ppa=unit)
		ACK(attach,ATTACH,(ackbuf))
	}

	/*
	** Bind (defer if using HP-UX 9 or HP-UX 10.20, totally skip if
	** using SINIX)
	*/
#if !defined(HAVE_HPUX9) && !defined(HAVE_HPUX10_20) && !defined(sinix)
#ifdef _AIX
	REQUEST2F(bind,BIND,sap=1537,service_mode=DL_CLDLS)
#else
#ifdef DL_HP_RAWDLS
ABORT THIS COMPILATION AND FIND WHY YOUR PLATFORM SETS DL_HP_RAWDLS
	/*
	 * This is another HPUX 'feature'. You will need to invent REQUEST3F to
	 * accomodate it.
	 */
#endif
	REQUEST2F(bind,BIND,sap=0,service_mode=DL_CLDLS)
#endif
	ACK(bind,BIND,(ackbuf))
#endif

	/* Do *NOT* set promiscuous here - all manner of strangeness
	 * will result because the interfaces will capture packets destined
	 * for other ethernet MACs.  (Because the code above does not
	 * check that the destination matches the device in use).
	 */

	/*
	 * ** Try to enable sap (when not in promiscuous mode, but never under
	 * SINIX)
	 */
#ifndef sinix
	REQUESTF(promiscon,PROMISCON,level=DL_PROMISC_SAP)
	ACK(promiscon,PROMISCON,(ackbuf))
#endif

	/*
	** HP-UX 9 and HP-UX 10.20 must bind after setting promiscuous
	** options)
	*/
#if defined(HAVE_HPUX9) || defined(HAVE_HPUX10_20)
	REQUEST2F(bind,BIND,sap=0,service_mode=DL_CLDLS)
	ACK(bind,BIND,(ackbuf))
#endif

#ifdef	DLIOCRAW
	/*
	** This is a non standard SunOS hack to get the full raw link-layer
	** header.
	*/
	if (ioctl(fd, DLIOCRAW, 0)<0)
		throw WireError(errno, "DLPI DLIOCRAW");
#endif

	/*
	 * Do not try to push bufmod here. The whole purpose of upslug2 is to
	 * send lots on data upstream, so it makes no sense to buffer stuff
	 * coming downstream. Moreover, bufmod comes with assorted 'gotchas'
	 * which are best avoided.
	 */

	if (ioctl(fd, I_FLUSH, FLUSHR) != 0)
		throw (errno, "FLUSHR");

	if (mac == NULL) {
		REQUESTF(phys_addr,PHYS_ADDR,addr_type=DL_CURR_PHYS_ADDR)
		ACK(phys_addr,PHYS_ADDR,(ackbuf))
		if (ackbuf->physaddr_ack.dl_addr_length != sizeof(macBuffer))
			throw WireError(0, "DLPI wrong length for local MAC address");
		std:memcpy(macBuffer, longbuf+ackbuf->physaddr_ack.dl_addr_offset,
				   sizeof(macBuffer));
		mac = macBuffer;
	}

	/* This is enough to make a new wire. */
	return new DlpiWire(fd, device, mac, address);
}
#endif
